#include <SparkFun_APDS9960.h>
#include <LiquidCrystal_I2C.h>
#include <SdFat.h>
#include <TMRpcm.h>

#define LCD_LEN             16
#define LCD_WIDTH           2

#define WRITE_START_BYTE    0x01
#define WRITE_END_BYTE      0x10
#define WRITE_FLAGS_SZ      8
#define TITLE_MAX           40
#define BUF_LEN             32
#define ACK                 "ack"

#define APDS9960_INT        2
#define SD_READER_CS        10
#define SPEAKER_PIN         9

#define PAUSE_BTN           PD3
#define NEXT_BTN            PD5
#define PREV_BTN            PD4
#define SHUFF_BTN           PD6
#define DEBOUNCE_DELAY      200

#define LED_PIN_1           A0
#define LED_PIN_2           PB0
#define LED_PIN_3           PD7

#define VOL_MAX             5
#define VOL_MIN             1
#define VOL_OFFSET          10

#define META_FILE         F("metadata")
#define LUCKY_SONG        "roulette.wav"
#define LUCKY_TIME         300

#define SONG_LEN_MAX        8

SparkFun_APDS9960 apds = SparkFun_APDS9960();
LiquidCrystal_I2C lcd(0x27, LCD_LEN, LCD_WIDTH);

// Id of the current song
volatile int songIdx;
volatile int songChanged = 0;

// Variable used as flag to catch gestures in APDS ISR
volatile int apds_gest_isr_flag = 0;

// Time counters
volatile bool secondPassed;
volatile bool halfSecondPassed;
volatile bool tenNsPassed;

// Storage which has to be accessible from anywhere within the code
SdFat SD;
File metadataFile;

TMRpcm music;
volatile int musicOn;
int songVolume;

volatile bool luckyMode;


ISR(INT0_vect) {
  apds_gest_isr_flag = 1;
}

ISR(INT1_vect) {
  static unsigned long midLastDebounceTime = 0;

  if ((millis() - midLastDebounceTime) > DEBOUNCE_DELAY) {
    midLastDebounceTime = millis();
    music.pause();
    musicOn ^= 1;
  }
}

ISR(PCINT2_vect) {
  static unsigned long PC0LastDebounceTime = 0;

  if ((millis() - PC0LastDebounceTime) > DEBOUNCE_DELAY) {
    PC0LastDebounceTime = millis();

    if ((PIND & (1 << NEXT_BTN)) == 0) {
      songIdx += 1;
      songChanged = 1;
    } else if ((PIND & (1 << PREV_BTN)) == 0) {
      songIdx -= 1;
      songChanged = 1;
    } else if ((PIND & (1 << SHUFF_BTN)) == 0) {
      luckyMode = 1;
      musicOn = 0;
    }
  }
}

ISR(TIMER2_COMPA_vect) {
  static unsigned int calls = 0;
  calls++;

  if (calls == 80) {
    calls = 0;
    secondPassed = 1;
  }

  if (calls % 40 == 0) {
    halfSecondPassed = 1;
  }

  tenNsPassed = 1;
}

static int getNumSongs() {
  char sz[4] = {0};

  metadataFile.seekSet(0);
  metadataFile.readBytesUntil('\n', sz, 4);

  return atoi(sz);
}

// Function that extracts the song name from the file containing
// all the filenames, based on the index and starts playing it
static void startSong(int index, char *songTitle, char *songLen) {
  PCMSK2 &= ~(1 << PCINT20) & ~(1 << PCINT21) & ~(1 << PCINT22);

  int szInt = getNumSongs();
  index = ((index % szInt) + szInt) % szInt;

  for (int i = 0; i < index; ++i)
    while (metadataFile.read() != '\n') ;

  memset(songTitle, 0, sizeof(songTitle) / sizeof(*songTitle));
  memset(songTitle, ' ', 2);
  size_t bytesRead = metadataFile.readBytesUntil('|', songTitle + 2, 63);
  (songTitle + 2)[bytesRead] = '\0';

  memset(songLen, 0, sizeof(songLen) / sizeof(*songLen));
  bytesRead = metadataFile.readBytesUntil('\n', songLen, 8);
  songLen[bytesRead - 1] = '\0';

  // Display the name of the song
  lcd.setCursor(0,0);
  lcd.print(songTitle);

  // Display the prefix for the total length of the song
  lcd.setCursor(0, 1);
  lcd.print(F("0:00/"));
  lcd.setCursor(5, 1);
  lcd.print(songLen);

  // Get the filename based on the id
  char songName[4] = {0};
  sprintf(songName, "%d.wav", index + 1);

  music.play(songName);
  displayVolume(songVolume);
  musicOn = 1;

  PCMSK2 |= (1 << PCINT20) | (1 << PCINT21) | (1 << PCINT22);
}

static void displayVolume(int volume) {
  lcd.setCursor(VOL_OFFSET, 1);
  lcd.print(F("|Vol:"));
  lcd.print(volume);
}

static void confTimer2() {
  TCCR2A = 0;
  TCCR2B = 0;
  TCNT2 = 0;

  // Compare on match register 16MHz/1024/625Hz - 1 => count to 24
  OCR2A = 24;

  // CTC mode
  TCCR2B |= (1 << WGM21);

  // 1024 prescaler: CS22, CS21, CS20 = 1, 1, 1
  TCCR2B |= (1 << CS22) | (1 << CS21) | (1 << CS20);
}

void interpretGesture() {
  // Deactivate interrupts on INT0 pin
  EIMSK &= ~(1 << INT0);

  if (apds.isGestureAvailable()) {
    switch (apds.readGesture()) {
      case DIR_UP:
        if (songVolume < VOL_MAX) {
          music.volume(1);
          songVolume++;
          displayVolume(songVolume);
        }

        break;
      case DIR_DOWN:
        if (songVolume > VOL_MIN) {
          music.volume(0);
          songVolume--;
          displayVolume(songVolume);
        }

        break;
      default:
        Serial.println("Invalid gesture");
    }
  }

  apds_gest_isr_flag = 0;

  // Reactivate INT0 interrupts
  EIMSK |= (1 << INT0);
}

static void luckyStartup() {
  music.disable();
  music.play(LUCKY_SONG);
  music.loop(true);
}

static void luckyCleanup() {
  lcd.clear();

  PORTB &= ~(1 << (LED_PIN_2 - PB0));
  PORTC &= ~(1 << (LED_PIN_1 - A0));
  PORTD &= ~(1 << LED_PIN_3);

  musicOn = 1;
  songIdx = random(SHRT_MAX);
  songChanged = 1;
}

void setup() {
  Serial.begin(115200);

  cli();
  DDRD |= (1 << LED_PIN_3);
  DDRB |= (1 << LED_PIN_2);

  DDRD &= ~(1 << NEXT_BTN) & ~(1 << PREV_BTN) & ~(1 << SHUFF_BTN) & ~(1 << PAUSE_BTN) & ~(1 << APDS9960_INT);
  PORTD |= (1 << NEXT_BTN) | (1 << PREV_BTN) | (1 << SHUFF_BTN) | (1 << PAUSE_BTN);

  // Activate external interrupts on both pins (2, 3)
  EIMSK |= (1 << INT0) | (1 << INT1);

  // Set INT0 and INT1 to trigger only on falling edge
  EICRA |= (1 << ISC01) | (1 << ISC11);

  // Activate Pin Change interrupts on PCINT20, PCINT21 and PCINT22
  PCMSK2 |= (1 << PCINT20) | (1 << PCINT21) | (1 << PCINT22);

  // Enable the Pin Change interrupt for PCINT 23..16
  PCICR |= (1 << PCIE2);

  // Enable timer2 compare interrupt
  confTimer2();
  TIMSK2 |= (1 << OCIE2A);

  pinMode(LED_PIN_1, OUTPUT);
  sei();

  // Enable the SD Card Reader
  if (!SD.begin(SD_READER_CS))
    Serial.println(F("initialization failed!"));

  // Enable the gesture sensor
  if (!apds.init())
    Serial.println(F("APDS-9960 initialization failed"));

  if (!apds.enableGestureSensor(true))
    Serial.println(F("Enabling the gesture sensor went wrong!"));
  else
    Serial.println(F("APDS-9960 successfully initialized!"));

  // Start the LCD
  lcd.init();
  lcd.clear();         
  lcd.backlight();
  lcd.setCursor(0, 0);
  lcd.print("Please press a");
  lcd.setCursor(0, 1);
  lcd.print("button to start");

  // Prepare the file containing all the filesystem metadata
  metadataFile = SD.open(META_FILE, FILE_READ);

  // Configure the music player
  songVolume = (VOL_MAX + VOL_MIN) / 2;
  music.speakerPin = SPEAKER_PIN;
  music.setVolume(songVolume);
  randomSeed(analogRead(1));
}

void loop() {
  // Internal counters used for events
  static unsigned int secondsCnt = 0;
  static unsigned int halfSecondsCnt = 0;
  static unsigned int luckyCnt = 0;

  // Data about the current file transfer
  static bool writeStarted = 0;
  static bool titleMode = 0;
  static int titlePos = 0;
  static int buffPos = 0;
  static char buff[BUF_LEN] = {0};
  static File currFile;

  // Data about the current song playing
  static char songTitle[TITLE_MAX];
  static char songLen[SONG_LEN_MAX];

  // Flags regarding the current state
  static bool luckyModeLoop = 0, luckyModeEnded = 0;

  if (musicOn && secondPassed) {
    secondPassed = 0;
    secondsCnt++;

    char displaySeconds[5] = {0};
    sprintf(displaySeconds, "%d:%02d", secondsCnt / 60, secondsCnt % 60);

    lcd.setCursor(0, 1);
    lcd.print(displaySeconds);

    // Transition to the next song if the current one ended
    if (!strcmp(songLen, displaySeconds))
      songChanged = 1;
  }

  // Shift the song title one position to the left on LCD
  if (musicOn && halfSecondPassed) {
    halfSecondPassed = 0;
    halfSecondsCnt++;

    for (int i = 0; i < LCD_LEN; ++i) {
      lcd.setCursor(i, 0);
      lcd.print(songTitle[(halfSecondsCnt + i) % strlen(songTitle)]);
    }
  }

  if (tenNsPassed && luckyModeLoop) {
    tenNsPassed = 0;
    luckyCnt++;

    int pos = luckyCnt % (LCD_WIDTH * LCD_LEN);
    lcd.setCursor(pos % LCD_LEN, pos / LCD_LEN);
    lcd.print((char)random(256));

    if (luckyCnt % 10 == 0) {
      if (luckyCnt % 3 == 1) {
        PORTC |= (1 << (LED_PIN_1 - A0));
        PORTD &= ~(1 << LED_PIN_3);
      } else if (luckyCnt % 3 == 2) {
        PORTD |= (1 << LED_PIN_3);
        PORTB &= ~(1 << (LED_PIN_2 - PB0));
      } else {
        PORTB |= (1 << (LED_PIN_2 - PB0));
        PORTC &= ~(1 << (LED_PIN_1 - A0));
      }
    }

    if (luckyCnt == LUCKY_TIME) {
      luckyCnt = 0;
      luckyModeEnded = 1;
    }
  }

  if (songChanged) {
    songChanged = 0;
    music.disable();
    secondsCnt = 0;
    halfSecondsCnt = 0;

    startSong(songIdx, songTitle, songLen);
  }

  if (luckyMode) {
    luckyMode = 0;
    luckyModeLoop = 1;
    luckyStartup();
  } else if (luckyModeEnded) {
    luckyModeEnded = 0;
    luckyModeLoop = 0;
    luckyCleanup();
  }

  // Adjust the volume according to the gesture input
  if(apds_gest_isr_flag == 1) {
    interpretGesture();
  }

  // Transfer mode (store data from laptop on SD Card)
  if (!musicOn && Serial.available() > 0) {
    music.disable();
    secondsCnt = 0;
    halfSecondsCnt = 0;

    char currByte = Serial.read();
    buff[buffPos++ % BUF_LEN] = currByte;

    // The next batch of BUF_LEN chars can be now received
    if (buffPos && (buffPos % BUF_LEN == 0))
      Serial.println(ACK);

    if (buffPos >= WRITE_FLAGS_SZ) {
      int eq_start = 1;
      int eq_end = 1;

      for (int i = 0; i < WRITE_FLAGS_SZ; ++i) {
        if (buff[(buffPos - 1 - i) % BUF_LEN] == WRITE_START_BYTE && !eq_start) {
          eq_end = 0;
          break;
        } else if (buff[(buffPos - 1 -i) % BUF_LEN] == WRITE_END_BYTE && !eq_end) {
          eq_start = 0;
          break;
        } else if (buff[(buffPos - 1 - i) % BUF_LEN] == WRITE_START_BYTE) {
          eq_end = 0;
        } else if (buff[(buffPos - 1 - i) % BUF_LEN] == WRITE_END_BYTE) {
          eq_start = 0;
        } else {
          eq_start = 0;
          eq_end = 0;
          break;
        }
      }

      if (eq_start) {
        writeStarted = 1;
        titleMode = titleMode ^ 1;

        if (!titleMode)
          currFile = SD.open(songTitle, FILE_WRITE);
      } else if (eq_end) {
        writeStarted = 0;

        if (!titleMode) {
          currFile.close();
        } else {
          songTitle[titlePos] = '\0';
          titlePos = 0;
        }

        buffPos = 0;
        Serial.println(songTitle);
      } else if (writeStarted) {
        if (buffPos >= 2 * WRITE_FLAGS_SZ)
          if (titleMode) {
            if (titlePos < 63)
              songTitle[titlePos++] = buff[(buffPos - 8) % BUF_LEN];
          } else {
             currFile.write(buff[(buffPos - 8) % BUF_LEN]);
          }
      }
    }
  }
}
